from orwynn.log import Log
from orwynn.websocket.context.id import (
    WebsocketRequestContextId,
)
from orwynn.websocket.middleware.builtin import (
    BuiltinWebsocketMiddleware,
)
from orwynn.websocket.middleware.nextcall import WebsocketNextCall
from orwynn.websocket.websocket import Websocket


class RequestContextBuiltinWebsocketMiddleware(BuiltinWebsocketMiddleware):
    """Populates the context storage with the current request's id."""
    async def process(
        self, request: Websocket, call_next: WebsocketNextCall
    ) -> None:
        request_id: str = WebsocketRequestContextId().save()

        # Also contextualize logs
        with Log.contextualize(**{"websocket.request_id": request_id}):
            return await call_next(request)



from typing import Callable, Sequence

from starlette.middleware.base import (
    BaseHTTPMiddleware as StarletteBaseHTTPMiddleware,
)

from orwynn.app import App
from orwynn.base.error.utils import get_exception_direct_subclasses
from orwynn.base.errorhandler import ErrorHandler
from orwynn.base.middleware import Middleware
from orwynn.http import (
    BUILTIN_HTTP_MIDDLEWARE,
    BuiltinHttpMiddleware,
    DefaultErrorHandler,
    DefaultHttpErrorHandler,
    ErrorHandlerHttpMiddleware,
    HttpMiddleware,
)
from orwynn.http.errors import HttpException
from orwynn.router.errorhandlermanager import ErrorHandlerManager
from orwynn.utils import validation
from orwynn.utils.scheme import Scheme
from orwynn.websocket import (
    BUILTIN_WEBSOCKET_MIDDLEWARE,
    BuiltinWebsocketMiddleware,
    ErrorHandlerWebsocketMiddleware,
    WebsocketMiddleware,
    WebsocketStack,
    routing_handlers,
)


class MiddlewareRegister:
    """
    Registers middleware to the system.

    Note that exception handlers is transforming into united middleware which
    handles all errors occured for a protocol.
    """
    def __init__(
        self,
        *,
        app: App,
        middleware_arr: list[Middleware],
        exception_handlers: list[ErrorHandler],
        websocket_stack: WebsocketStack,
    ) -> None:
        self.__app: App = app

        self.__middleware_arr: list[Middleware] = middleware_arr
        self.__exception_handlers: set[ErrorHandler] = set(
            exception_handlers
        )

        self.__websocket_stack: WebsocketStack = websocket_stack
        self.__is_websocket_middleware_added: bool = False

    def register_all(self) -> None:
        """
        Registers all middleware to the system.
        """
        populated_handlers_py_protocol: dict[
            Scheme, set[ErrorHandler]
        ] = ErrorHandlerManager().get_populated_handlers_by_protocol(
            self.__exception_handlers
        )

        # FIXME: Here below the builtin middleware are created directly without
        #   the Di notifying which may raise a confusion in the next calls.
        #   Also it might become harder to support if an builtin middleware
        #   requests any Di-related dependency.

        http_builtin_middleware: Sequence[
            BuiltinHttpMiddleware
        ] = self.__collect_http_builtin_middleware(
            populated_handlers_py_protocol[Scheme.HTTP]
        )
        websocket_builtin_middleware: Sequence[
            BuiltinWebsocketMiddleware
        ] = self.__collect_websocket_builtin_middleware(
            populated_handlers_py_protocol[Scheme.WEBSOCKET]
        )

        self.__register_middleware_arr(
            # Add builtin middlewares first, and others second. No matters the
            # order between middleware of different protocols.
            http_builtin_middleware
            + websocket_builtin_middleware
            + self.__middleware_arr
        )

    def __collect_http_builtin_middleware(
        self,
        handlers: set[ErrorHandler]
    ) -> list[BuiltinHttpMiddleware]:
        middleware_arr: list[BuiltinHttpMiddleware] = []

        for Middleware_ in BUILTIN_HTTP_MIDDLEWARE:
            if issubclass(Middleware_, ErrorHandlerHttpMiddleware):
                middleware_arr.append(Middleware_(
                    handlers
                ))
            else:
                middleware_arr.append(Middleware_())

        return middleware_arr

    def __collect_websocket_builtin_middleware(
        self,
        handlers: set[ErrorHandler]
    ) -> list[BuiltinWebsocketMiddleware]:
        middleware_arr: list[BuiltinWebsocketMiddleware] = []

        for Middleware_ in BUILTIN_WEBSOCKET_MIDDLEWARE:
            if issubclass(
                Middleware_, ErrorHandlerWebsocketMiddleware
            ):
                middleware_arr.append(Middleware_(
                    handlers
                ))
            else:
                middleware_arr.append(Middleware_())

        return middleware_arr

    def __register_middleware_arr(
        self,
        middleware_arr: Sequence[Middleware],
    ) -> None:
        """
        Adds middleware to the system.

        Args:
            middleware_arr:
                Middleware list to be added to the system.
            cors:
                Cors object (can be None) to configure the Cors middleware on
                the fly.
        """
        # Note that middleware here is reversed since Starlette.add_middleware
        # inserts new functions at the top of the middleware list which makes
        # older added middlewares executable last, which is not logical.
        #
        # But this reversion should be executed only for HtppMiddleware since
        # for Websockets we have own logic.
        #
        # So, first task is to separate middleware
        http_middleware_arr: list[HttpMiddleware] = []
        websocket_middleware_arr: list[WebsocketMiddleware] = []

        for middleware in middleware_arr:
            if isinstance(middleware, HttpMiddleware):
                http_middleware_arr.append(middleware)
            elif isinstance(middleware, WebsocketMiddleware):
                websocket_middleware_arr.append(middleware)
            elif type(middleware) is Middleware:
                raise TypeError(
                    f"cannot register an instance {middleware} of an abstact"
                    " class Middleware"
                )
            else:
                raise TypeError(
                    f"unrecognized middleware {middleware}"
                )

        # Add HTTP from reversed list to comply Starlette
        for http_middleware in reversed(http_middleware_arr):
            self.__register_middleware_unit(http_middleware)

        # Add Websocket normally
        if websocket_middleware_arr != []:
            if self.__is_websocket_middleware_added:
                raise ValueError(
                    "websocket middleware have been already added"
                )

            for websocket_middleware in websocket_middleware_arr:
                self.__register_middleware_unit(websocket_middleware)

            self.__is_websocket_middleware_added = True

    def __register_middleware_unit(self, middleware: Middleware) -> None:
        validation.validate(middleware, Middleware)

        # Note that dispatch(...) method is linked to be as entrypoint to a
        # middleware. This will be a place where a middleware takes decision
        # to not process request to certain endpoint or not.
        if type(middleware) is Middleware:
            raise TypeError(
                f"cannot accept abstract class implementation {middleware}"
            )
        elif isinstance(middleware, ErrorHandlerHttpMiddleware):
            self.__register_exception_http_middleware(middleware)
        elif isinstance(middleware, HttpMiddleware):
            self.__register_http_middleware_fn(
                middleware.dispatch
            )
        elif isinstance(middleware, WebsocketMiddleware):
            self.__websocket_stack.add_call(
                routing_handlers.DispatchWebsocketHandler(
                    fn=middleware.dispatch
                )
            )
        else:
            raise TypeError(
                f"unrecognized middleware {middleware}"
            )

    def __register_http_middleware_fn(
        self,
        fn: Callable
    ) -> None:
        self.__app._fw_register_middleware(
            StarletteBaseHTTPMiddleware,
            dispatch=fn
        )

    def __register_exception_http_middleware(
        self,
        middleware: ErrorHandlerHttpMiddleware
    ) -> None:
        # NOTE: It may seem strange that firstly exception handlers are wrapped
        #   into middleware and then unwrapped here for HTTP protocol, but
        #   in this way we comply with other protocols (such as Websocket).
        #   Adding middleware directly as BaseHTTPMiddleware is not an option
        #   since HTTPException (such as 404 Not Found) won't be handled.
        __RemainingExceptionDirectSubclasses: set[type[Exception]] = \
            set(get_exception_direct_subclasses())

        custom_base_exception_handler: ErrorHandler | None = None

        for handler in middleware.handlers:
            __RemainingExceptionDirectSubclasses.discard(
                handler.HandledException
            )

            if handler.HandledException is Exception:
                # Do not add the base exception explicitly since Starlette
                # cannot handle it, add all it's direct subclasses instead
                custom_base_exception_handler = handler
                continue
            else:
                self.__app._fw_register_exception_handler_fn(
                    handler.HandledException,
                    handler._fw_handle_wrapper
                )

        for Remaining in __RemainingExceptionDirectSubclasses:
            handle_fn: Callable
            if Remaining is HttpException:
                handle_fn = DefaultHttpErrorHandler().handle
            else:
                # If custom exception handler is defined, pass it's handler,
                # for all remaining subclasses, else use default exc handler
                if custom_base_exception_handler:
                    handle_fn = \
                        custom_base_exception_handler._fw_handle_wrapper
                else:
                    handle_fn = DefaultErrorHandler()._fw_handle_wrapper

            self.__app._fw_register_exception_handler_fn(
                Remaining,
                handle_fn
            )



from typing import Any, Self

from sqlalchemy import ForeignKey
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from orwynn.utils.fmt import snakefy
from orwynn.utils.rnd import makeid


class Table(DeclarativeBase):
    """Base orm model responsible of holding database model's data and
    it's basic CRUD operations.

    Update representatives are defined individually at each subclass
    (e.g. `set_something()`), and by default accessed via basic model
    alteration, e.g. `MyModel.name = 'Another name'`.
    """
    # Do not map Table to actual database table
    __abstract__ = True

    # For each table special string id is generated, but it is not a primary
    # key for the performance sake, for details see:
    #   https://stackoverflow.com/a/517591/14748231
    _id: Mapped[str] = mapped_column(default=makeid, unique=True)

    @hybrid_property
    def sid(self) -> int:
        return self._sid

    @declared_attr.cascading  # type: ignore
    def _sid(cls) -> Mapped[int]:
        parent: type[Self] = cls._get_parent_class()
        if parent is Table:
            return mapped_column(primary_key=True)
        else:
            return mapped_column(
                ForeignKey(snakefy(parent.__name__) + "._sid"),
                primary_key=True
            )

    # Declaring _type as class atrribute might be a reason for sqlalchemy to
    # raise a warning, maybe because this field is being redefined by
    # subclasses.
    @declared_attr
    def _type(self) -> Mapped[str]:
        return mapped_column()

    @classmethod
    def _get_parent_class(cls) -> type[Self]:
        bases: tuple[type] = cls.__bases__

        if len(bases) == 1:
            return bases[0]
        elif len(bases) > 1:
            raise ValueError(
                "multiple inheritance for Tables is not supported"
            )
        else:
            raise

    @hybrid_property
    def type(self) -> str:
        return self._type

    @hybrid_property
    def id(self) -> str:
        return self._id

    @declared_attr.directive
    def __tablename__(cls) -> str:
        cls_name: str = cls.__name__
        return snakefy(cls_name)

    @declared_attr.directive
    def __identity__(cls) -> str:
        return snakefy(cls.__name__)

    @declared_attr.directive
    def __mapper_args__(cls) -> dict[str, Any]:
        args: dict[str, Any] = {}
        args.update({
            "polymorphic_on": "_type",
            "polymorphic_identity": cls.__identity__
        })
        return args



class InvalidIdError(Exception):
    """
    Cannot convert string id to object id.
    """
    def __init__(
        self,
        *,
        invalid_id: str
    ) -> None:
        message: str = f"cannot convert id <{invalid_id}> to ObjectId"
        super().__init__(message)


class InvalidOperatorError(Exception):
    """
    MongoDB operator is invalid.
    """
    def __init__(
        self,
        *,
        operator: str,
        explanation: str
    ):
        message: str = f"cannot accept operator <{operator}>: {explanation}"
        super().__init__(message)



from orwynn.utils.mp import patch
from orwynn.utils.mp.dictpp import dictpp


def test_patch():
    a = dictpp({
        "BurgerShot": {
            "location": "Vinewood",
            "employees": 15,
            "menu": {
                "standard_burger": 1.25,
                "big_burger": 2.3,
                "cola": 1.5
            }
        }
    })
    b = dictpp({
        "BurgerShot": {
            "menu": {
                "cola": 1.8,
                "pizza": 4.1,
                "donut": 1.3
            }
        }
    })

    result = patch(a, b)

    assert result == {
        "BurgerShot": {
            "location": "Vinewood",
            "employees": 15,
            "menu": {
                "standard_burger": 1.25,
                "big_burger": 2.3,
                "cola": 1.8,
                "pizza": 4.1,
                "donut": 1.3
            }
        }
    }


def test_patch_no_deepcopy():
    a = dictpp({
        "BurgerShot": {
            "location": "Vinewood",
            "employees": 15,
            "menu": {
                "standard_burger": 1.25,
                "big_burger": 2.3,
                "cola": 1.5
            }
        }
    })
    b = dictpp({
        "BurgerShot": {
            "menu": {
                "cola": 1.8,
                "pizza": 4.1,
                "donut": 1.3
            }
        }
    })

    patch(a, b, should_deepcopy=False)

    assert a == {
        "BurgerShot": {
            "location": "Vinewood",
            "employees": 15,
            "menu": {
                "standard_burger": 1.25,
                "big_burger": 2.3,
                "cola": 1.8,
                "pizza": 4.1,
                "donut": 1.3
            }
        }
    }



import inspect
from types import NoneType

from orwynn.base.error.errors import MalfunctionError
from orwynn.base.error.utils import (
    find_detailed_class_for_exception,
)
from orwynn.base.errorhandler.errorhandler import ErrorHandler
from orwynn.utils import validation
from orwynn.websocket.context.id import (
    WebsocketRequestContextId,
)
from orwynn.websocket.log.logger import WebsocketLogger
from orwynn.websocket.middleware.builtin import (
    BuiltinWebsocketMiddleware,
)
from orwynn.websocket.middleware.nextcall import WebsocketNextCall
from orwynn.websocket.websocket import Websocket


class ErrorHandlerWebsocketMiddleware(BuiltinWebsocketMiddleware):
    """
    Handles all errors occured at Websocket layer.
    """
    def __init__(
        self,
        handlers: set[ErrorHandler]
    ) -> None:
        super().__init__()
        validation.validate_each(
            handlers, ErrorHandler, expected_sequence_type=set
        )
        self.__handlers: set[ErrorHandler] = handlers

    @property
    def handlers(self) -> set[ErrorHandler]:
        return self.__handlers.copy()

    async def process(
        self,
        request: Websocket,
        call_next: WebsocketNextCall
    ) -> None:
        try:
            await call_next(request)
        except Exception as err:  # noqa: BLE001
            is_handled: bool = False
            malfunction_message: str = \
                "malfunction at ErrorHandler middleware" \
                " for a websocket connection: no handler to handle an" \
                f" error {err}"

            # Choose according handler
            AppropriateException: type[Exception] = \
                find_detailed_class_for_exception(
                    err,
                    [handler.HandledException for handler in self.__handlers]
                )
            for handler in self.__handlers:
                if handler.HandledException is AppropriateException:
                    if not isinstance(err, handler.HandledException):
                        malfunction_message = \
                            f"appropriate Exception {AppropriateException}" \
                            f" is not a parent for actual exception {err}"
                    elif not inspect.iscoroutinefunction(handler.handle):
                        malfunction_message = \
                            f"handler function {handler.handle} is not a" \
                            " coroutine"
                    else:
                        validation.apply(
                            await handler.handle(request, err),
                            NoneType
                        )
                        is_handled = True

                    # Finally exit after the first match
                    break

            if not is_handled:
                # Not recommended to raise an error since it won't be
                # logged properly at this layer, better to explicitly log and
                # digest the error on the fly.
                data: dict = MalfunctionError(malfunction_message).api
                await WebsocketLogger().log_response(
                    data,
                    request=request,
                    request_id=WebsocketRequestContextId().get()
                )
                await request.send_json(
                    data
                )

            # Close the request after handling (even on malfunction error),
            # otherwise it will be stuck
            await request.close()



from pathlib import Path
from typing import Optional, Sequence
from orwynn.utils import validation
from orwynn.base.database.database import Database

from .table import Table

from .databasekind import SQLDatabaseKind

from orwynn.sql.config import SqlConfig
from sqlalchemy import create_engine, Engine
from sqlalchemy import Table as SQLAlchemyTable
from sqlalchemy.orm import Session


class Sql(Database):
    def __init__(
        self,
        config: SqlConfig
    ) -> None:
        super().__init__()
        self.__config = config
        self.__engine: Engine = self._create_engine()

    @property
    def engine(self) -> Engine:
        return self.__engine

    @property
    def session(self) -> Session:
        """Returns a new session."""
        return Session(self.__engine)

    def create_tables(self, *tables: type[Table]) -> None:
        """Creates tables.

        Args:
            *tables:
                Tables to create. If no tables are given (or None), all
                reachable tables are created.
        """
        Table.metadata.create_all(
            self.__engine,
            tables=self._collect_sqla_tables(*tables)
        )

    def drop_tables(self, *tables: type[Table]) -> None:
        """Drops tables.

        Args:
            *tables:
                Tables to drop. If no tables are given (or None), all
                reachable tables are dropped.
        """
        Table.metadata.drop_all(
            self.__engine,
            tables=self._collect_sqla_tables(*tables)
        )

    def _collect_sqla_tables(
        self,
        *tables: type[Table]
    ) -> Optional[Sequence[SQLAlchemyTable]]:
        sqla_tables: Optional[Sequence[SQLAlchemyTable]] = None
        if tables:
            sqla_tables = []
            for table in tables:
                sqla_tables.append(table.__table__)  # type: ignore
        return sqla_tables

    def _create_engine(self) -> Engine:
        url: str
        connect_args: dict = {}
        final_kwargs: dict = {}

        match self.__config.database_kind:
            case SQLDatabaseKind.POSTGRESQL:
                url = \
                    f"postgresql://{self.__config.database_user}" \
                    + f":{self.__config.database_password}" \
                    + f"@{self.__config.database_host}" \
                    + f":{self.__config.database_port}" \
                    + f"/{self.__config.database_name}"
            case SQLDatabaseKind.SQLITE:
                database_path: str = validation.apply(
                    self.__config.database_path,
                    str
                )

                # Create directories for the path if they are not exist.
                # https://stackoverflow.com/a/273227/14748231
                Path(database_path).parent.mkdir(parents=True, exist_ok=True)

                # Note that here we add additional slash since in either
                # memory ":memory:", relative "var/app.db" or absolute
                # "/tmp/app.db" cases we have to prepend slash to comply with
                # SQLite standard.
                url = f"sqlite:///{database_path}"

                # Make SQLite know that more than one thread could interact
                # with the database for the same request.
                # See https://fastapi.tiangolo.com/tutorial/sql-databases/#note
                connect_args.update({"check_same_thread": False})
            case _:
                raise TypeError(
                    f"unknown database {self.__config.database_kind}"
                )

        final_kwargs["connect_args"] = connect_args
        final_kwargs["poolclass"] = self.__config.real_poolclass
        if self.__config.pool_size:
            final_kwargs["pool_size"] = self.__config.pool_size

        return create_engine(
            url,
            **final_kwargs
        )



from orwynn.http.context.id import HttpRequestContextId
from orwynn.http.log.configs import LogHttpMiddlewareConfig
from orwynn.http.log.logger import HttpLogger
from orwynn.http.middleware.middleware import HttpMiddleware
from orwynn.http.middleware.nextcall import HttpNextCall
from orwynn.http.requests import HttpRequest
from orwynn.http.responses import HttpResponse


class LogMiddleware(HttpMiddleware):
    """Logs information about request and response lihking them together.

    It's recommended to be outermost (at custom level) middleware.
    """
    def __init__(
        self, covered_routes: list[str], config: LogHttpMiddlewareConfig
    ) -> None:
        super().__init__(covered_routes)

        self.__http_logger: HttpLogger = HttpLogger()
        self._config: LogHttpMiddlewareConfig = config

    async def process(
        self, request: HttpRequest, call_next: HttpNextCall
    ) -> HttpResponse:
        request_id: str = HttpRequestContextId().get()

        if self._config.is_request_logged:
            await self.__http_logger.log_request(
                request,
                request_id
            )

        response: HttpResponse = await call_next(request)

        if self._config.is_response_logged:
            await self.__http_logger.log_response(
                response,
                request=request,
                request_id=request_id
            )

        return response



from orwynn.base.module.module import Module
from orwynn.http import (
    Endpoint,
    HttpController,
    HttpMiddleware,
    HttpNextCall,
    HttpRequest,
    HttpResponse,
)
from orwynn.utils import validation


def test_controller_added_to_no_route():
    class C1(HttpController):
        ROUTE = "/"
        ENDPOINTS = [
            Endpoint(
                method="get"
            )
        ]

        def get(self, *args, **kwargs) -> dict:
            return {}

    validation.expect(
        Module,
        ValueError,
        route=None,
        Controllers=[C1]
    )


def test_middleware_added_to_no_route():
    class M1(HttpMiddleware):
        async def process(
            self, request: HttpRequest, call_next: HttpNextCall
        ) -> HttpResponse:
            return await super().process(request, call_next)

    validation.expect(
        Module,
        ValueError,
        route=None,
        Middleware=[M1]
    )



from enum import Enum


class Validator(Enum):
    """Conventional object to represent cases requiring special validation.

    Values:
        SKIP:
            Don't care about type, just skip
    """
    SKIP = 0



"""Common types.
"""
from typing import Any, Callable, Coroutine, TypeVar

CommonType = TypeVar("CommonType")
Class = TypeVar("Class", bound=type)
DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])

Timestamp = float
Delta = float

CashOperator = str
"""
Special string literal starts with dollar sign `$` which holds a special
meaning to the acceptor logic.
"""

AnyCoro = Coroutine[Any, Any, Any]



from orwynn.base.middleware import Middleware
from orwynn.http.middleware.nextcall import HttpNextCall
from orwynn.http.requests import HttpRequest
from orwynn.http.responses import HttpResponse


class HttpMiddleware(Middleware):
    """Intermediate operational layer for HTTP requests."""
    async def dispatch(
        self,
        request: HttpRequest,
        call_next: HttpNextCall
    ) -> HttpResponse:
        return await super().dispatch(request, call_next)

    async def process(
        self,
        request: HttpRequest,
        call_next: HttpNextCall
    ) -> HttpResponse:
        return await super().process(request, call_next)



from collections.abc import Awaitable
from typing import Callable

from orwynn.http.requests import HttpRequest
from orwynn.http.responses import HttpResponse

# Callable function which returns an awaitable, see:
#   https://stackoverflow.com/a/59177557/14748231
HttpNextCall = Callable[[HttpRequest], Awaitable[HttpResponse]]



import pytest

from orwynn.mongo.document import Document


class SimpleDocument(Document):
    name: str
    price: float
    priority: int = 5


class NestedDocument(Document):
    nested: dict


@pytest.fixture
def document_1(mongo_boot) -> SimpleDocument:
    return SimpleDocument(
        name="pizza",
        price=1.2,
        priority=2
    ).create()


@pytest.fixture
def document_2(mongo_boot) -> SimpleDocument:
    return SimpleDocument(
        name="donut",
        price=1
    ).create()


@pytest.fixture
def nested_document_1(mongo_boot) -> NestedDocument:
    return NestedDocument(
        nested={
            "key1": {
                "key2": 1
            }
        }
    ).create()



import json
from types import NoneType
from typing import Literal

import pytest

from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.http import Endpoint, HttpController, LogMiddleware
from orwynn.log import Log
from orwynn.proxy.boot import BootProxy
from orwynn.testing import Writer
from orwynn.testing.client import Client


def _check_log_message(message: str) -> list[dict]:
    assert message != ""
    items: list[str] = message.split("\n")
    parsed_items: list[dict] = []

    for item in items:
        if item == "":
            continue
        data: dict = json.loads(str(item))
        extra: dict = data["record"]["extra"]

        try:
            http_extra: dict = extra["http"]
        except KeyError as error:
            raise AssertionError from error

        request_or_response: Literal["request", "response"]
        if "request" in http_extra and "response" not in http_extra:
            request_or_response = "request"
            request_data: dict = http_extra["request"]
            assert type(request_data["id"]) is str
            assert type(request_data["url"]) is str
        elif "response" in http_extra and "request" not in http_extra:
            request_or_response = "response"
            response_data: dict = http_extra["response"]
            assert type(response_data["request_id"]) is str
            assert type(response_data["status_code"]) is int
            assert type(response_data["media_type"]) in [str, NoneType]
        else:
            raise AssertionError

        assert type(http_extra[request_or_response]["headers"]) is dict
        assert http_extra[request_or_response]["json"] is None or (
            # Empty json dict is not allowed in log extra - it should be
            # replaced with None.
            type(http_extra[request_or_response]["json"]) is dict
            and extra[request_or_response]["json"] != {}
        )

        parsed_items.append(data)

    return parsed_items


@pytest.mark.asyncio
async def test_get(
    writer: Writer,
    log_apprc_sink_to_writer: dict
):
    RETURNED_DATA: dict = {"hello": 1}

    class C1(HttpController):
        ROUTE = "/"
        ENDPOINTS = [Endpoint(method="get")]

        def get(self) -> dict:
            return RETURNED_DATA

    boot: Boot = await Boot.create(
        Module(
            route="/", Controllers=[C1], Middleware=[LogMiddleware]
        ),
        apprc=log_apprc_sink_to_writer
    )  # type: ignore #worker
    client: Client = boot.app.client
    client.get(
        "/",
        200
    )

    items: list[dict] = _check_log_message(writer.read())
    for item in items:
        extra: dict = item["record"]["extra"]
        if "request" in extra:
            assert extra["http.request"]["json"] is None
        elif "response" in extra:
            assert extra["http.response"]["json"] == RETURNED_DATA

    Log.remove()


@pytest.mark.asyncio
async def test_get__error(
    writer: Writer,
    log_apprc_sink_to_writer: dict
):
    class C1(HttpController):
        ROUTE = "/"
        ENDPOINTS = [Endpoint(method="get")]

        def get(self) -> dict:
            raise ValueError("hello")

    boot: Boot = await Boot.create(
        Module(
            route="/", Controllers=[C1], Middleware=[LogMiddleware]
        ),
        apprc=log_apprc_sink_to_writer
    )  # type: ignore #worker
    client: Client = boot.app.client
    client.get(
        "/",
        400
    )

    items: list[dict] = _check_log_message(writer.read())
    for item in items:
        extra: dict = item["record"]["extra"]
        if "request" in extra:
            assert extra["http.request"]["json"] is None
        elif "response" in extra:
            assert \
                extra["http.response"]["json"] \
                == BootProxy.ie().api_indication.digest(ValueError("hello"))

    Log.remove()



from unittest.mock import patch

from fastapi.routing import APIRouter as _NativeApiRouter

from orwynn.app.apiwebsocketroute import ApiWebsocketRoute


class ApiRouter(_NativeApiRouter):
    def add_api_websocket_route(self, *args, **kwargs) -> None:
        with patch("fastapi.routing.APIWebSocketRoute", new=ApiWebsocketRoute):
            return super().add_api_websocket_route(*args, **kwargs)



from enum import Enum


class IndicationType(Enum):
    """Possible values of field "type".

    Values:
        OK:
            Everything is correct.
        ERROR:
            Some error is happened either by client's problem or by server's
            issue. Here goes also all errors connected with business-logic
            itself.
    """
    OK = "ok"
    # FIXME: In future, consider moving all non-BL related problems
    # to another type, e.g. "http-error".
    ERROR = "error"



from unittest.mock import patch

from fastapi import FastAPI
from fastapi.middleware.asyncexitstack import AsyncExitStackMiddleware
from starlette.middleware import Middleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.middleware.exceptions import ExceptionMiddleware
from starlette.types import ASGIApp

from orwynn.app.apirouter import ApiRouter


class CoreApp(FastAPI):
    def __init__(self, **kwargs) -> None:
        with patch("fastapi.routing.APIRouter", new=ApiRouter):
            super().__init__(**kwargs)

    def build_middleware_stack(self) -> ASGIApp:
        # Duplicate/override from Starlette to add AsyncExitStackMiddleware
        # inside of ExceptionMiddleware, inside of custom user middlewares
        debug = self.debug
        error_handler = None
        exception_handlers = {}

        for key, value in self.exception_handlers.items():
            if key in (500, Exception):
                error_handler = value
            else:
                exception_handlers[key] = value

        middleware = [
            Middleware(
            ServerErrorMiddleware, handler=error_handler, debug=debug
            ),
            Middleware(
                ExceptionMiddleware,
                handlers=exception_handlers,
                debug=debug
            ),
            *self.user_middleware,
            Middleware(AsyncExitStackMiddleware)
        ]

        app = self.router
        for cls, options in reversed(middleware):
            app = cls(app=app, **options)
        return app



import pytest

from orwynn.base.module import Module
from orwynn.base.service import Service
from orwynn.boot import Boot
from orwynn.bootscript.bootscript import Bootscript
from orwynn.bootscript.calltime import CallTime
from orwynn.di.di import Di
from orwynn.utils import validation


class SomeService(Service):
    def __init__(self) -> None:
        super().__init__()
        self.some_var: int = 0


def fn(some_service: SomeService) -> None:
    some_service.some_var = 1


async def async_fn(some_service: SomeService) -> None:
    some_service.some_var = 5


@pytest.mark.asyncio
async def test_basic():
    await Boot.create(
        Module(Providers=[SomeService]),
        bootscripts=[
            Bootscript(
                fn=fn,
                call_time=CallTime.AFTER_ALL
            )
        ]
    )

    some_service: SomeService = validation.apply(
        Di.ie().find("SomeService"),
        SomeService
    )

    assert some_service.some_var == 1


@pytest.mark.asyncio
async def test_async():
    await Boot.create(
        Module(Providers=[SomeService]),
        bootscripts=[
            Bootscript(
                fn=fn,
                call_time=CallTime.AFTER_ALL
            )
        ]
    )

    some_service: SomeService = validation.apply(
        Di.ie().find("SomeService"),
        SomeService
    )

    assert some_service.some_var == 1



"""List of classes an user can subclass from."""
from orwynn.base.controller.controller import Controller
from orwynn.base.errorhandler.errorhandler import ErrorHandler
from orwynn.base.middleware import Middleware
from orwynn.base.model.model import Model
from orwynn.base.service.service import Service
from orwynn.utils.klass.types import BaseSubclassable

# Note that here listed the most basic classes. E.g. Config is not listed
# since it is a derivative from the Model and on the stage of DI
# it will be checked to find these more specific classes.
SUBCLASSABLES: list[BaseSubclassable] = [
    Service,
    Controller,
    Middleware,
    Model,
    ErrorHandler
]

ENVIRONS: set[str] = {
    "ORWYNN_MODE",
    "ORWYNN_ROOT_DIR",
    "ORWYNN_APPRC_PATH",
    "ORWYNN_IS_CATCH_LOGGING_ENABLED_IN_TESTS"
}
"""
All environment variables the framework supports.
"""



from orwynn.base.model import Model
from orwynn.bootscript.callable import BootscriptCallable
from orwynn.bootscript.calltime import CallTime


class Bootscript(Model):
    fn: BootscriptCallable
    call_time: CallTime

    class Config:
        arbitrary_types_allowed = True




from orwynn.base.controller.controller import Controller
from orwynn.base.middleware import GlobalMiddlewareSetup, Middleware
from orwynn.base.model.model import Model
from orwynn.base.module.module import Module
from orwynn.di.collecting.acceptordependencies import (
    collect_dependencies_for_acceptor,
)
from orwynn.di.container import DiContainer
from orwynn.http import HttpController, HttpMiddleware
from orwynn.proxy.boot import BootProxy
from orwynn.utils import validation
from orwynn.utils.url import join_routes
from orwynn.utils.validation import validate
from orwynn.websocket import WebsocketController, WebsocketMiddleware


class __CoveredRoutes(Model):
    http: list[str]
    websocket: list[str]


__ModuleToCoveredRoutes = tuple[Module, __CoveredRoutes]


def init_other_acceptors(
    container: DiContainer,
    modules: list[Module],
    global_middleware: GlobalMiddlewareSetup | None = None
) -> None:
    """Populates DI container with initialized non-provider acceptors.

    Attributes:
        container:
            DI container.
        modules:
            List of modules to collect acceptors from.
        global_middleware (optional):
            Global middleware setup setting. None is initialized by default.
    """
    if global_middleware is None:
        global_middleware = {}
    __init_modules(container, modules, global_middleware)


def __init_modules(
    container: DiContainer,
    modules: list[Module],
    global_middleware: GlobalMiddlewareSetup
) -> None:
    # Init global middleware first, and then normally module's defined.
    __init_middleware(
        container=container,
        source=global_middleware
    )

    _init_all_error_handlers(container)

    for module in modules:
        covered_routes: __CoveredRoutes = __init_controllers(container, module)
        __init_middleware(
            container=container,
            source=(module, covered_routes)
        )


def _init_all_error_handlers(
    container: DiContainer
) -> None:
    for Eh in BootProxy.ie().ErrorHandlers:
        container.add(
            Eh(**collect_dependencies_for_acceptor(
                Eh,
                container,
                None
            ))
        )


def __init_controllers(
    container: DiContainer,
    module: Module
) -> __CoveredRoutes:
    http_covered_routes: list[str] = []
    websocket_covered_routes: list[str] = []

    for C in module.Controllers:
        validate(C, Controller)

        if module.route is None:
            raise ValueError(
                f"module {module} has not route and shouldn't have added"
                " any controllers or middleware"
            )

        controller: Controller = C(
            **collect_dependencies_for_acceptor(
                C,
                container,
                module
            )
        )
        final_route: str = join_routes(
            module.route, controller.route
        )

        container.add(
            controller
        )

        if issubclass(C, HttpController):
            http_covered_routes.append(final_route)
        elif issubclass(C, WebsocketController):
            # The final route should be duplicated for each controller's
            # event handler, since they are forming the controller's
            # subroute system
            for subroute in C.get_handler_subroutes():
                websocket_covered_routes.append(
                    join_routes(final_route, subroute)
                )
        elif type(C) is Controller:
            raise TypeError(
                f"abstract Controller instance {C} is disallowed"
            )
        else:
            raise TypeError(
                f"unrecognized controller {C}"
            )

    return __CoveredRoutes(
        http=http_covered_routes,
        websocket=websocket_covered_routes
    )


def __init_middleware(
    container: DiContainer,
    source: __ModuleToCoveredRoutes | GlobalMiddlewareSetup
) -> None:
    """
    Inits a middleware.

    Args:
        container:
        source:
            Module and it's covered routes or global middleware setting.
    """
    if isinstance(source, tuple):
        validation.validate_length(source, 2)
        __init_middleware_from_module(
            container=container,
            module=source[0],
            covered_routes=source[1]
        )
    elif isinstance(source, dict):
        __init_global_middleware(
            container=container,
            setup=source
        )
    else:
        raise TypeError(
            f"unrecognized source object {source}"
        )


def __init_global_middleware(
    container: DiContainer,
    setup: GlobalMiddlewareSetup
) -> None:
    for Middleware_, covered_routes_arr in setup.items():
        __register_middleware_in_container(
            container=container,
            Middleware_=Middleware_,
            covered_routes_arr=covered_routes_arr,
            # Omit the availability check
            module=None
        )


def __init_middleware_from_module(
    *,
    container: DiContainer,
    module: Module,
    covered_routes: __CoveredRoutes
) -> None:
    for Mw in module.Middleware:
        validate(Mw, Middleware)

        if module.route is None:
            raise ValueError(
                f"module {module} has not route and shouldn't have added"
                " any controllers or middleware"
            )

        final_covered_routes: list[str]
        if issubclass(Mw, HttpMiddleware):
            final_covered_routes = covered_routes.http
        elif issubclass(Mw, WebsocketMiddleware):
            final_covered_routes = covered_routes.websocket
        elif type(Mw) is Middleware:
            raise TypeError(
                f"abstract Middleware instance {Mw} is disallowed"
            )
        else:
            raise TypeError(
                f"unrecognized middleware {Mw}"
            )

        __register_middleware_in_container(
            container=container,
            Middleware_=Mw,
            covered_routes_arr=final_covered_routes,
            module=module
        )


def __register_middleware_in_container(
    *,
    container: DiContainer,
    Middleware_: type[Middleware],
    covered_routes_arr: list[str],
    module: Module | None
) -> None:
    """
    Adds a middleware to the container.

    Args:
        ...
        module:
            Module to check dependencies availability from. If None, all
            initialized dependencies will be available.
    """
    container.add(
        Middleware_(
            covered_routes=covered_routes_arr,
            **collect_dependencies_for_acceptor(
                Middleware_,
                container,
                module
            )
        )
    )



from orwynn.http.middleware.middleware import HttpMiddleware


class BuiltinHttpMiddleware(HttpMiddleware):
    """Special framework's middleware which covers all possible routes."""
    def __init__(self) -> None:
        # Cover all routes
        super().__init__(["*"])



from orwynn.utils.url.helpers import get_vars
from orwynn.utils.url.url import Url
from orwynn.utils.url.vars import UrlVars


def test_path():
    vars: UrlVars = get_vars(
        url=Url("/user/eg1"),
        abstract_route="/user/{user_id}"
    )
    assert vars.path_vars["user_id"] == "eg1"
    assert vars.query_vars == {}


def test_path_2():
    vars: UrlVars = get_vars(
        url=Url("/user/eg1/hello"),
        abstract_route="/user/{user_id}/{user_sign}"
    )
    assert vars.path_vars["user_id"] == "eg1"
    assert vars.path_vars["user_sign"] == "hello"
    assert vars.query_vars == {}


def test_query():
    vars: UrlVars = get_vars(
        url=Url("/user?token=woo"),
        abstract_route="/user"
    )
    assert vars.path_vars == {}
    assert vars.query_vars["token"] == "woo"


def test_query_2():
    vars: UrlVars = get_vars(
        url=Url("/user?token=woo&order=5"),
        abstract_route="/user"
    )
    assert vars.path_vars == {}
    assert vars.query_vars["token"] == "woo"
    assert vars.query_vars["order"] == "5"


def test_mixed():
    vars: UrlVars = get_vars(
        url=Url("/user/eg1/hello?token=woo&order=5"),
        abstract_route="/user/{user_id}/{user_sign}"
    )
    assert vars.path_vars["user_id"] == "eg1"
    assert vars.path_vars["user_sign"] == "hello"
    assert vars.query_vars["token"] == "woo"
    assert vars.query_vars["order"] == "5"



"""
Collection of validation functions.
"""
import re
import typing
from inspect import isclass
from pathlib import Path
from typing import Any, Callable, Optional, Sized, TypeVar

from pydantic import ValidationError as _PydanticValidationError
from pydantic import validator as _pydantic_validator
from orwynn.utils.types import AnyCoro

from orwynn.utils.validation.validator import Validator
from orwynn.utils.validation.errors import (ExpectationError,
                                            ReValidationError,
                                            UnknownValidatorError,
                                            ValidationError)

# WARNING: typing aliases are not currently supported so passing types like
#   "dict[str, Any]" to check will produce ValidationError in any case since
#   direct runtime comparison is made.
ValidationExpectedType = type | list[type] | Validator | Path
ApplyExpectedType = TypeVar("ApplyExpectedType")
model_validator = _pydantic_validator
ModelValidationError = _PydanticValidationError


def validate(
    obj: Any,
    expected_type: ValidationExpectedType,
    *,
    is_strict: bool = False
) -> None:
    """Validates given object against expected type.

    Args:
        obj:
            Object to be validated.
        expected_type:
            Type to compare object to.
        is_strict (optional):
            Whether strict check should be performed. If True, direct type
            comparison is made, disallowing subclasses. If False, isinstance()
            comparison is made.
    Raises:
        ValidationError:
            Object did not pass validation.
    """
    if isinstance(expected_type, Validator):
        __check_validator(obj, expected_type)
    elif expected_type is Callable:
        __check_callable(obj, typing.cast(Callable, expected_type), is_strict)
    elif isinstance(expected_type, type):
        __check_type(obj, expected_type, is_strict)
    elif type(expected_type) is list:
        is_matched_type_found: bool = False

        for type_ in expected_type:
            try:
                validate(obj, type_, is_strict=is_strict)
            except ValidationError:
                continue
            else:
                is_matched_type_found = True

        if not is_matched_type_found:
            raise ValidationError(
                failed_obj=obj, expected_type=expected_type
            )
    else:
        raise ValidationError(
            f"{expected_type} should be Type, an instance of list or"
            " Validator"
        )


def validate_each(
    obj: list | tuple | set | frozenset,
    expected_type: ValidationExpectedType,
    *,
    is_strict: bool = False,
    expected_sequence_type: type | None = None,
    should_check_if_empty: bool = False
) -> None:
    """Validates each object in given array against expected type.

    Args:
        obj:
            Object to be validated. Note that not all iterables are accepted,
            only meaningful ones. E.g. there is not sence for this function
            to accept string and validate each char in it.
        expected_type:
            Type to compare object to.
        is_strict (optional):
            Whether strict check should be performed. If True, direct type
            comparison is made, disallowing subclasses. If False, isinstance()
            comparison is made.
        expected_sequence_type (optional):
            To perform checking for the given sequence itself.
        should_check_if_empty (optional):
            Perform checking if given obj is empty.
    """
    validate(obj, [list, tuple, set, frozenset], is_strict=True)

    if expected_sequence_type is not None:
        if expected_sequence_type not in [list, tuple, set, frozenset]:
            raise ValidationError(
                f"expected object type {expected_sequence_type} is neither"
                " list, tuple, set or frozen set"
            )
        validate(obj, expected_sequence_type)

    is_empty: bool = True
    for o in obj:
        is_empty = False
        validate(o, expected_type, is_strict=is_strict)

    if should_check_if_empty and is_empty:
        raise ValidationError("validated iterable shouldn't be empty")


def validate_dict(
    obj: dict,
    expected_types: tuple[ValidationExpectedType, ValidationExpectedType],
    *,
    strict_flags: tuple[bool, bool] | None = None
) -> None:
    """Validates each key and each value in given dict.

    Args:
        obj:
            Dict to be validated.
        expected_types:
            Tuple of expected types for key and expected types for value.
        strict_flags (optional):
            Tuple of flags for strict validation separately for key and for
            value. Defaults to (False, False) which means no strict validation
            neither for key nor for value.

    Raises:
        NothingToValidateError:
            Empty dict given.
        ValidationError:
            Some key or value did not passed the validation.
    """
    if strict_flags is None:
        strict_flags = (False, False)

    validate(obj, dict)
    validate_length(strict_flags, 2)
    validate_each(strict_flags, bool)

    for k, v in obj.items():
        validate(k, expected_types[0], is_strict=strict_flags[0])
        validate(v, expected_types[1], is_strict=strict_flags[1])


def validate_re(string: str, pattern: str) -> None:
    """Validates given string using re.match.

    Args:
        string:
            String to validate.
        pattern:
            Regex pattern to apply.

    Raises:
        ReValidationError:
            String does not match given pattern.
    """
    if not re.match(pattern, string):
        raise ReValidationError(failed_obj=string, pattern=pattern)


def validate_length(obj: Sized, expected_length: int) -> None:
    """Validates length of the given object.

    Args:
        obj:
            Sized object to be validated.
        expected_length:
            Which length is expected.

    Raise:
        ValidationError:
            Given object is not Sized.
        ValidationError:
            Length expectations failed.
    """
    if not hasattr(obj, "__len__"):
        raise ValidationError(
            f"object {obj} should implement \"__len__\" method"
        )

    obj_len: int = len(obj)
    if not obj_len == expected_length:
        raise ValidationError(
            f"sized object {obj} length {obj_len} != "
            f" expected length {expected_length}"
        )


def validate_route(route: str) -> None:
    """Validates route.

    Args:
        route:
            Route to validate.

    Raises:
        ReValidationError:
            Route does not match route pattern.
    """
    validate_re(route, r"^\/(.+\/?)?$")


def apply(
    obj: Any, expected_type: type[ApplyExpectedType]
) -> ApplyExpectedType:
    """Validates given object against given expected type.

    Args:
        obj:
            Object to be validated.
        expected_type:
            Type to check.

    Returns:
        The same object but with given type reference.

    Raises:
        ValidationError:
            Object type is not expected type.
    """
    validate(obj, expected_type)
    return obj


def expect(
    fn: Callable,
    ErrorToExpect: type[Exception],
    *args,
    **kwargs
) -> None:
    """
    Expects given function to raise given error if function is called with
    given args and kwargs.

    Args:
        fn:
            Function to call.
        ErrorToExpect:
            Exception class to expect.
        args:
            Positional arguments to pass to function call.
        kwargs:
            Keyword arguments to pass to function call.

    Raises:
        ExpectationError:
            Given error has not been raised on function's call.
    """
    try:
        fn(*args, **kwargs)
    except ErrorToExpect:
        pass
    else:
        raise ExpectationError(
            f"error {ErrorToExpect} expected on call of function {fn}"
        )


async def expect_async(
    coro: AnyCoro,
    ErrorToExpect: type[Exception]
) -> None:
    """
    Works same as validation.expect, but with async coroutine being tested.
    """
    try:
        await coro
    except ErrorToExpect:
        pass
    else:
        raise ExpectationError(
            f"error {ErrorToExpect} expected on call of coro {coro}"
        )


CheckedObj = TypeVar("CheckedObj")
def check(obj: Optional[CheckedObj]) -> CheckedObj:
    """Checks if given object is not None.

    Raises:
        ValidationError:
            If given object is None.
    """
    if obj is None:
        raise ValidationError(f"shouldn't be None")
    return obj


def __check_validator(obj: Any, validator: Validator) -> None:
    match validator:
        case Validator.SKIP:
            return
        case _:
            raise UnknownValidatorError(
                f"unknown validator {validator}"
            )


def __check_callable(obj: Any, c: Callable, is_strict: bool) -> None:
    if is_strict:
        raise ValueError(
            "expected type is Callable and strict flag is true"
            " which is not logical"
        )
    elif not callable(obj):
        raise ValidationError(
            f"{obj} is not Callable"
        )


def __check_type(obj: Any, t: type, is_strict: bool) -> None:
    if is_strict and type(obj) is t:
        return
    elif isinstance(obj, t):
        return
    elif isclass(obj) and issubclass(obj, t):
        return

    raise ValidationError(
        failed_obj=obj, expected_type=t
    )



from orwynn.base.model.model import Model
from orwynn.cli.enums import CliCommand


class CliCommandData(Model):
    """
    Data required for the chosen mode.

    @abstract
    """


class RunCliCommandData(CliCommandData):
    """
    Data sufficient for modes: Test, Dev and Prod.
    """
    host: str
    port: int


class CliArgs(Model):
    command: CliCommand
    data: CliCommandData



import pytest
import starlette.websockets

from orwynn.apiversion import ApiVersion
from orwynn.base.module.module import Module
from orwynn.boot import Boot
from orwynn.websocket.controller.controller import (
    WebsocketController,
)
from orwynn.websocket.websocket import Websocket


class WsCtrl(WebsocketController):
    ROUTE = "/message"

    async def main(self, ws: Websocket) -> None:
        await ws.send_json({"message": "hello"})


@pytest.mark.asyncio
async def test_default():
    """
    By default a client should use the global route.
    """
    boot: Boot = await Boot.create(
        root_module=Module("/", Controllers=[WsCtrl]),
        global_websocket_route="/donuts"
    )

    with boot.app.client.websocket("/message") as ws:
        data: dict = ws.receive_json()
        assert data["message"] == "hello"


@pytest.mark.asyncio
async def test_default_version():
    """
    By default a client should use the latest api version available.
    """
    class _C(WsCtrl):
        VERSION = 3

        async def main(self, ws: Websocket) -> None:
            return await super().main(ws)

    boot: Boot = await Boot.create(
        root_module=Module("/", Controllers=[_C]),
        global_websocket_route="/ws/v{version}",
        api_version=ApiVersion(
            supported={1, 2, 3}
        )
    )

    with boot.app.client.websocket("/message") as ws:
        data: dict = ws.receive_json()
        assert data["message"] == "hello"


@pytest.mark.asyncio
async def test_not_used():
    """
    The global route can be disabled.
    """
    boot: Boot = await Boot.create(
        root_module=Module("/", Controllers=[WsCtrl]),
        global_websocket_route="/donuts"
    )

    with boot.app.client.websocket(
        "/donuts/message",
        is_global_route_used=False
    ) as ws:
        data: dict = ws.receive_json()
        assert data["message"] == "hello"


@pytest.mark.asyncio
async def test_pass_version():
    """
    A client is able to not specify global route, but pass own api version.
    """
    class _C(WsCtrl):
        VERSION = 2

        async def main(self, ws: Websocket) -> None:
            return await super().main(ws)

    boot: Boot = await Boot.create(
        root_module=Module("/", Controllers=[_C]),
        global_websocket_route="/ws/v{version}",
        api_version=ApiVersion(
            supported={1, 2, 3}
        )
    )

    try:
        with boot.app.client.websocket(
            "/message",
            api_version=1
        ) as ws:
            pass
    except starlette.websockets.WebSocketDisconnect:
        pass
    else:
        raise AssertionError()

    with boot.app.client.websocket(
        "/message",
        api_version=2
    ) as ws:
        data: dict = ws.receive_json()
        assert data["message"] == "hello"

    try:
        with boot.app.client.websocket(
            "/message",
            api_version=3
        ) as ws:
            pass
    except starlette.websockets.WebSocketDisconnect:
        pass
    else:
        raise AssertionError()



import re
from typing import TYPE_CHECKING, Any, Self

from orwynn.base.model.model import Model
from orwynn.proxy.boot import BootProxy
from orwynn.utils import validation

if TYPE_CHECKING:
    from orwynn.apprc.apprc import AppRc


class Config(Model):
    """Object holding configuration which can be injected to any requesting
    entity.
    """
    @classmethod
    def load(cls, *, extra: dict[str, Any] | None = None) -> Self:
        if not extra:
            extra = {}

        validation.validate_dict(extra, (str, validation.Validator.SKIP))

        apprc: "AppRc" = BootProxy.ie().apprc
        config_kwargs: dict[str, Any] = apprc.get(
            cls._convert_name_to_rc_format(),
            {}
        )
        return cls(**validation.apply(config_kwargs, dict), **extra)

    @classmethod
    def _convert_name_to_rc_format(cls) -> str:
        cleaned_name: str = cls.__name__
        if re.match(r"^.+config$", cls.__name__.lower()):
            cleaned_name = cls.__name__[:len(cls.__name__) - 6]
        return cleaned_name

    class Config:
        arbitrary_types_allowed = True



from orwynn.log import Log
from orwynn.utils import validation
from orwynn.websocket.websocket import Websocket


class WebsocketLogger:
    """Logs websocket requests."""
    async def log_request(
        self,
        request: Websocket,
        request_id: str
    ) -> str:
        """Assigns special id to request and logs it.

        Args:
            request:
                Request to be logged.
            request_id:
                ID of the request.

        Returns:
            Request assigned ID.
        """
        validation.validate(request, Websocket)
        validation.validate(request_id, str)

        plain_message: str = \
            f"websocket CONNECT {request.url.path}" \
            f"{request.url.query}"

        extra: dict = {
            "websocket.request": {
                "id": request_id,
                # Get full URL
                "url": request.url._url,
                "headers": dict(request.headers)
            }
        }

        Log.bind(**extra).info(
            plain_message
        )

        return request_id

    async def log_response(
        self,
        response_data: dict,
        *,
        request: Websocket,
        request_id: str
    ) -> None:
        """
        Logs a response linking it to the according request.
        """
        plain_message: str = \
            "websocket response" \
            f" {request.url.path}{request.url.query}:" \
            f" {response_data}"

        extra: dict = {
            "websocket.response": {
                "request_id": request_id,
                "json": response_data
            }
        }

        Log.bind(**extra).info(
            plain_message
        )



from typing import Any

# Location is several nested dicrionary keys joined with ".", e.g. if you have
# dictionary:
# ```python
# ```
# location of field with value 10 will be "a1.a2.a3"
FieldLocation = str


class FieldNotFoundError(Exception):
    pass


class MalformedLocationError(Exception):
    pass


def find_location_by_field(field: Any, mp: dict) -> FieldLocation:
    """Searches given dictionary to find given field.

    Args:
        field:
            Field value to search.
        mp:
            Dictionary to search in. All keys should be parseable to str.

    Returns:
        Location of field in form "key1.key2.key3" relative to dictionary root.

    Raises:
        FieldNotFoundError:
            No key with given field found in map.
    """
    key_chain: list[str] = []

    is_found: bool = __traverse_comparing_field(field, mp, key_chain)

    if not is_found:
        raise FieldNotFoundError(
            f"no key with field {field} in given map {mp}"
        )

    key_chain.reverse()
    return ".".join(key_chain)


def find_field_by_location(location: FieldLocation, mp: dict) -> Any:
    """Searches given map to find location and extract field.

    Args:
        location:
            What search for.
        mp:
            What to search in.

    Returns:
        Field value found.
    """
    result: Any = mp
    keys: list[str] = __get_location_keys(location)
    keys.reverse()

    while keys:
        try:
            result = result[keys.pop()]
        except KeyError as err:
            raise FieldNotFoundError(
                f"no field for location \"{location}\" in map {mp}"
            ) from err

    return result


def __get_location_keys(location: FieldLocation) -> list[str]:
    if location == "":
        raise MalformedLocationError("empty location given")

    keys: list[str] = location.split(".")

    if keys == []:
        raise MalformedLocationError(
            f"malformed location {location}"
        )

    return keys


def __traverse_comparing_field(
    field: Any,
    mp: dict,
    key_chain: list[str]
) -> bool:
    # Returns flag signifies whether a key was appended.

    for k, v in mp.items():
        if (
            (
                isinstance(v, dict)
                and __traverse_comparing_field(field, v, key_chain)
            )
            or v == field
        ):
            key_chain.append(str(k))
            return True

    return False



from pydantic import BaseModel

from orwynn.utils.url.url import Url


class UrlVars(BaseModel):
    url: Url
    abstract_route: str
    path_vars: dict[str, str]
    query_vars: dict[str, str]

    class Config:
        arbitrary_types_allowed = True



import functools
from typing import Any, Callable

from orwynn.utils.types import Class

from .errors import ClassNotFoundError


class Static:
    def __init__(self) -> None:
        raise NotImplementedError


def find_subclass_by_name(name: str, BaseClass: Class) -> Class:
    """Searches given base class for subclass with given name.

    Args:
        name:
            Name of the target class.
        BaseClass:
            Where to search for subclasses.

    Returns:
        Class found.
    """
    # Maybe given class has correct name
    if BaseClass.__name__ == name:
        return BaseClass

    out: Class | None = __traverse_subclasses_checking_name(name, BaseClass)

    if out is None:
        raise ClassNotFoundError(
            f"class of supertype {BaseClass} with name {name} is not found"
        )
    else:
        return out

def bind_first_arg(arg: Any):
    """Adds first argument to wrapped function call.

    Useful when you want to add e.g. "self" with some instance to dynamically
    obtained class's method.

    Args:
        arg:
            First arg to be added to wrapped function call.

    Returns:
        Wrapper accepting function.
    """
    def wrapper(fn: Callable):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            return fn(arg, *args, **kwargs)
        return inner
    return wrapper


def bind_first_arg_async(arg: Any):
    """Same as bind_first_arg(), but for async functions.
    """
    def wrapper(fn: Callable):
        @functools.wraps(fn)
        async def inner(*args, **kwargs):
            return await fn(arg, *args, **kwargs)
        return inner
    return wrapper


def __traverse_subclasses_checking_name(name: str, C: Class) -> Class | None:
    for SubClass in C.__subclasses__():
        if SubClass.__name__ == name:
            return SubClass
        else:
            result: Class | None = \
                __traverse_subclasses_checking_name(name, SubClass)
            if result is not None:
                return result

    return None






class DigestingError(Exception):
    pass


class UnsupportedIndicatorError(Exception):
    pass


class RecoveringError(Exception):
    """Wasn't able to recover an object from dictionary using indication."""



import argparse
import sys

from orwynn.cli.enums import CliCommand
from orwynn.cli.models import CliArgs, CliCommandData, RunCliCommandData


class Cli:
    """
    Manages incoming CLI input and calls appropriate Orwynn commands.
    """
    def execute(
        self,
        args: list[str] | None = None
    ) -> None:
        """
        Executes arguments as a commands to Orwynn.

        Args:
            args(optional):
                List of arguments splitted by space. Defaults to system input.
        """


    def _parse_args(
        self,
        args: list[str] | None = None
    ) -> CliArgs:
        arg_list: list[str]
        if args:
            arg_list = args
        else:
            arg_list = sys.argv

        # TODO(ryzhovalex): ensure call like "python -m ..." does not broke
        #   this
        # 0
        command: CliCommand = CliCommand(arg_list[2])

        return CliArgs(
            command=command,
            data=self._get_cli_command_data(
                command,
                arg_list[3:]
            )
        )

    def _get_cli_command_data(
        self,
        command: CliCommand,
        data_args: list[str]
    ) -> CliCommandData:
        data: CliCommandData
        namespace: argparse.Namespace
        parser: argparse.ArgumentParser = argparse.ArgumentParser(
            description=f"Orwynn {command.value.capitalize()} CLI"
        )

        match command:
            case CliCommand.Test:
                raise NotImplementedError
            case CliCommand.Dev:
                self._add_dev_args(parser)
                namespace = parser.parse_args(data_args)
                host, port =
                data = RunCliCommandData(
                    host=namespace.
                )
            case CliCommand.Prod:
                self._add_prod_args(parser)


        return data

    def _add_dev_args(self, parser: argparse.ArgumentParser) -> None:
        self._add_run_args(parser)

    def _add_prod_args(self, parser: argparse.ArgumentParser) -> None:
        self._add_run_args(parser)

    def _add_run_args(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "address",
            type=str,
            help="address to serve on in format 'host:port'"
        )



import inspect

from orwynn.base.config import Config
from orwynn.base.model.model import Model
from orwynn.di.collecting.errors import (
    ProviderKeywordAttributeError,
)
from orwynn.di.errors import NoAnnotationError, NotProviderError
from orwynn.di.isprovider import is_provider
from orwynn.di.provider import Provider

ProviderParameters = list["ProviderParameter"]


class ProviderParameter(Model):
    name: str
    DependencyProvider: type[Provider]


def get_parameters_for_provider(
    P: type[Provider]
) -> ProviderParameters:
    # Inspects the provider and returns requested by it parameters.

    parameters: ProviderParameters = []

    for inspect_parameter in inspect.signature(P).parameters.values():
        if inspect_parameter.annotation is inspect._empty:
            raise NoAnnotationError(
                f"provider {P} has field \"{inspect_parameter.name}\" without"
                " annotation"
            )

        if (
            inspect_parameter.kind is inspect._ParameterKind.KEYWORD_ONLY
            and not issubclass(P, Config)
        ):
            raise ProviderKeywordAttributeError(
                f"provider {P} cannot have keyword only attributes"
            )

        # Comply to Liskov's principle - don't raise an error for *args and
        # **kwargs to be substitutable with base classes.
        # And yes, i haven't found any way to check if requested argument is
        # an either *positional or **keyword spreading, so conventional names
        # are checked.
        if inspect_parameter.name in ["args", "kwargs"]:
            continue

        # Note that on this stage all config parameters (even not providers) is
        # added, and later on additional checks is performed. Actually this
        # doesn't have much sense to not filter such non-provider parameters
        # here, and in future it might be refactored. But from other side it
        # might be useful later on to perform additional logic in DI's scope
        # on Config's non-provider fields, such as checking if it is not
        # waiting for other providers but i'm not sure that it's the case.

        if type(inspect_parameter.annotation) is str:
            raise NotImplementedError(
                "future string references,"
                f" like \"{inspect_parameter.annotation}\""
                f" in field \"{inspect_parameter.name}\""
                f" of provider {P}"
                " are not supported for now"
            )

        if (
            not inspect.isclass(inspect_parameter.annotation)
            or not is_provider(inspect_parameter.annotation)
        ):
            if issubclass(P, Config):
                continue
            else:
                raise NotProviderError(
                    f"argument annotation {inspect_parameter.annotation}"
                    f" for provider {P} should be provider"
                )

        parameters.append(
            ProviderParameter(
                name=inspect_parameter.name,
                DependencyProvider=inspect_parameter.annotation
            )
        )

    return parameters



"""Operation with yml.
"""
from enum import Enum
from pathlib import Path
from typing import Any

import yaml

from orwynn.utils.validation import validate


class YmlLoader(Enum):
    """Yaml loaders types according to
    https://github.com/yaml/pyyaml/wiki/PyYAML-yaml.load(input)-Deprecation
    """
    BASE = yaml.SafeLoader
    SAFE = yaml.FullLoader
    FULL = yaml.BaseLoader
    UNSAFE = yaml.UnsafeLoader


class NotValidYmlError(Exception):
    """If loaded yml is not valid."""


class NotValidFileSuffixError(Exception):
    pass


def load_yml(
    p: Path, *, loader: YmlLoader = YmlLoader.SAFE
) -> dict[str, Any]:
    """Loads yaml from file.

    Args:
        p:
            Path of yaml file to load from.
        loader (optional):
            Chosen loader for yaml. Defaults to safe loader.

    Returns:
        Loaded dictionary from yaml file.

    Raise:
        TypeError:
            Given path is not allowed pathlib.Path kind.
        NotValidFileSuffixError:
            Suffix should be either ".yml" or ".yaml".
        NotValidYmlError:
            Yaml file is not valid.
    """
    validate(p, Path)

    if p.suffix.lower() not in [".yaml", ".yml"]:
        raise NotValidFileSuffixError(f"suffix {p.suffix} is not valid suffix")

    with open(p) as file:
        data = yaml.load(file, Loader=loader.value)  # noqa: S506
        if data is None:
            # Empty files should return empty dicts
            data = {}
        # Is it necessary? Does pyyaml allow loading not-valid yaml files?
        elif type(data) is not dict:
            raise NotValidYmlError(
                "Yaml file should contain any map-like structure,"
                " not plain types"
            )

    return data



import copy
from types import NoneType
from typing import Any, Iterable, Self

from bson import ObjectId
from pydantic.fields import ModelField
from pymongo.cursor import Cursor
from pymongo.errors import DuplicateKeyError as PymongoDuplicateKeyError

from orwynn.di.di import Di
from orwynn.helpers.errors import UnsupportedError
from orwynn.mapping.errors import CustomUseOfMappingReservedFieldError
from orwynn.mapping.mapping import Mapping, if_linked
from orwynn.mongo.clientsession import ClientSession
from orwynn.mongo.document.errors import InvalidOperatorError
from orwynn.mongo.document.helpers import convert_to_object_id
from orwynn.mongo.entity import MongoEntity
from orwynn.mongo.errors import DocumentUpdateError, DuplicateKeyError
from orwynn.mongo.mongo import Mongo
from orwynn.utils import validation
from orwynn.utils.fmt import snakefy


class Document(Mapping):
    """
    Mapping to work with MongoDB.

    Itself is some model representing MongoDB document and also has some class
    methods to manipulate with related document in DB and translate it from/to
    mapping.

    The ID of the document on creating is always a string, not ObjectId for
    adjusting convenience. Under the hood a convertation str->ObjectId is
    performed before saving to MongoDB and backwards ObjectId->str before
    forming the document from MongoDB data.
    """
    def __init__(self, **data: Any) -> None:
        for k in data:
            if k.startswith("mongo_"):
                raise CustomUseOfMappingReservedFieldError(
                    f"field {k} for mapping {self.__class__} is reserved,"
                    " your field keys shouldn't be prefixed with \"mongo_\""
                )
        super().__init__(**data)

    @classmethod
    def start_session(cls, **kwargs) -> ClientSession:
        # Tip: In future here you can add per-document class defined session
        #   options to apply, so @classmethod is used instead of @staticmethod
        return cls._get_mongo().start_session(**kwargs)

    @classmethod
    def get(
        cls,
        query: dict | None = None,
        **kwargs
    ) -> Iterable[Self]:
        """
        Fetches all instances matching the query for this document.

        Args:
            query(optional):
                MongoDB-compliant dictionary to search. By default all
                instances of the document is fetched.
            **kwargs(optional):
                Additional arguments to Mongo's find method.

        Returns:
            Iterable with the results of the search.
        """
        return cls.find_all(query, **kwargs)

    @classmethod
    def find_all(
        cls,
        query: dict | None = None,
        **kwargs
    ) -> Iterable[Self]:
        """
        @deprecated use Document.get instead
        """
        _query: dict = cls._parse_query(query)
        validation.validate(_query, dict)

        cursor: Cursor = cls._get_mongo().find_all(
            cls._get_collection(),
            cls._adjust_id_to_mongo(_query),
            **kwargs
        )

        return map(cls._parse_document, cursor)

    @classmethod
    def find_one(
        cls,
        query: dict | None = None,
        **kwargs
    ) -> Self:
        """
        @deprecated use Document.get instead and select as many instances as
        you might need
        """
        _query: dict = cls._parse_query(query)
        validation.validate(_query, dict)

        return cls._parse_document(
            cls._get_mongo().find_one(
                cls._get_collection(),
                cls._adjust_id_to_mongo(_query),
                **kwargs
            )
        )

    @if_linked
    def create(
        self,
        session: ClientSession | None = None
    ) -> Self:
        data: dict = self._adjust_id_to_mongo(self.dict())

        try:
            return self._parse_document(
                self._get_mongo().create_one(
                    self._get_collection(), data, session=session
                )
            )
        except PymongoDuplicateKeyError as error:
            raise DuplicateKeyError(original_error=error) from error

    @if_linked
    def remove(
        self, **kwargs
    ) -> Self:
        id: str = self.getid()
        return self._parse_document(
            self._get_mongo().remove_one(
                self._get_collection(), {"_id": ObjectId(id)},
                **kwargs
            )
        )

    @if_linked
    def update(
        self,
        *,
        set: dict | None = None,
        inc: dict | None = None,
        operators: dict | None = None,
        **kwargs
    ) -> Self:
        """Updates document with given data.

        Args:
            set (optional):
                Which fields to set.
            inc (optional):
                Which fields to increment.
            operators(optional):
                Additional operators besides `$set` and `$inc` to add.
        """
        # Optimization tip: Consider adapting $inc in future for appropriate
        #   cases
        validation.validate(set, [dict, NoneType])
        validation.validate(inc, [dict, NoneType])

        id: str = self.getid()

        operation: dict = {}
        if set is not None:
            operation["$set"] = set
        if inc is not None:
            operation["$inc"] = inc
        if operators is not None:
            if "$set" in operators:
                raise InvalidOperatorError(
                    operator="$set",
                    explanation="pass it via keyword argument `set=`"
                )
            elif "$inc" in operators:
                raise InvalidOperatorError(
                    operator="$inc",
                    explanation="pass it via keyword argument `inc=`"
                )
            else:
                operation.update(operators)

        return self._parse_document(
            self._get_mongo().update_one(
                self._get_collection(),
                {"_id": ObjectId(id)},
                operation,
                **kwargs
            )
        )

    def refresh(
        self
    ) -> Self:
        """
        Refreshes the document with a new data from the database.
        """
        return self.find_one({
            "id": self.getid()
        })

    @classmethod
    def _get_collection(cls) -> str:
        return snakefy(cls.__name__)

    @classmethod
    def _get_mongo(cls) -> Mongo:
        return validation.apply(Di.ie().find("Mongo"), Mongo)

    @classmethod
    def _parse_document(cls, document: MongoEntity) -> Self:
        """Parses document to specified Model."""
        return cls.parse_obj(cls._adjust_id_from_mongo(document))

    @staticmethod
    def _adjust_id_to_mongo(data: dict) -> dict:
        if "id" in data:
            input_id_value: Any = data["id"]
            if input_id_value is not None:
                if (
                    isinstance(input_id_value, (str, dict, list))
                ):
                    data["_id"] = convert_to_object_id(input_id_value)
                else:
                    raise UnsupportedError(
                        title="field \"id\" with value",
                        value=input_id_value
                    )
            del data["id"]
        return data

    @staticmethod
    def _parse_query(query: dict | None) -> dict:
        return {} if query is None else copy.copy(query)

    @staticmethod
    def _adjust_id_from_mongo(data: dict) -> dict:
        if "_id" in data:
            if data["_id"] is not None:
                data["id"] = str(data["_id"])
            del data["_id"]
        return data



from orwynn.base.module.module import Module

from .config import SqlConfig
from .sql import Sql
from .table import Table

module = Module(
    Providers=[Sql, SqlConfig],
    exports=[Sql, SqlConfig]
)



from typing import Any


class UnsupportedError(Exception):
    """
    Some value is not recozniged/supported by the system.
    """
    def __init__(
        self,
        *,
        title: str | None = None,
        value: Any
    ) -> None:
        message: str

        pre_title: str = ""
        if title is not None:
            pre_title = f"{title}="

        message = f"unsupported value {pre_title}{value}"

        super().__init__(message)


class DeprecatedFeatureError(Exception):
    """
    Feature is deprecated.
    """
    def __init__(
        self,
        *,
        deprecated_feature: str,
        use_instead: str
    ):
        message: str = \
            f"feature <{deprecated_feature}> is deprecated," \
            f" use {use_instead} instead"
        super().__init__(message)



from orwynn.mongo.document.testing import SimpleDocument


def test_main(
    document_1: SimpleDocument
):
    document_1.update(set={
        "price": 5.0
    })

    assert document_1.price == 1.2

    document_1 = document_1.refresh()

    assert document_1.price == 5.0



from enum import Enum


class ServerEngine(Enum):
    """
    The server used under the hood.

    Attributes:
        Uvicorn:
            ASGI web server implementation for Python.
    """
    Uvicorn = "uvicorn"



from enum import Enum
import json
from typing import Any, ItemsView

from orwynn.base.error import get_error_code, ErrorCode
from orwynn.base.error import ErrorValueSchema
from orwynn.base.model.model import Model
from orwynn.http.schema.value import \
    HttpExceptionValueSchema
from orwynn.http.schema.validationvalue import \
    RequestValidationExceptionValueSchema
from orwynn.http.errors import HttpException
from orwynn.indication.type import IndicationType
from orwynn.indication.errors import UnsupportedIndicatorError
from orwynn.utils import validation
from orwynn.utils.mp.location import (FieldLocation, find_field_by_location,
                                     find_location_by_field)
from orwynn.utils.validation import (validate,
                                    validate_dict)
from orwynn.utils.validation.errors import RequestValidationException
from orwynn.utils.validation.validator import Validator

from .indicatable import Indicatable, IndicatableTypeVar
from .indicator import Indicator

_Locations = dict[Indicator, FieldLocation]


class Indication:
    """Holds some representation of how the data should be built like to serve
    as an convention for building objects.

    Each indication is simply mapping between strings and Indicators, but also
    implements special digest_* methods to transform given object applying
    defined indication rules.

    Also has counter-methods recover_* to create object back from dictionary.

    Strings-only as keys are used for an additional constraints for objects
    building an API structures, in future it is possible to see here an
    additional flags controlling this behaviour.

    For the list of applicable types see Indicator.
    """
    def __init__(
        self,
        mp: dict[str, Indicator]
    ) -> None:
        # FIXME: Temporarily indicators should be placed at highest dictionary
        #   level
        validate_dict(
            mp, (str, Indicator)
        )

        self.__mp: dict[str, Indicator] = mp
        self.__locations: _Locations = self.__find_locations()
        self.__created_schema_by_name: dict[str, type[Model]] = {}

    def gen_schema(
        self,
        Entity: type[IndicatableTypeVar]
    ) -> type[Model]:
        schema_name: str = Entity.__name__ + "IndicationSchema"
        schema_kwargs: dict[str, Any] = {}

        for k, v in self.items:
            final_field: str | type
            match v:
                case Indicator.TYPE:
                    final_field = str
                case Indicator.VALUE:
                    if issubclass(Entity, HttpException):
                        final_field = HttpExceptionValueSchema
                    elif issubclass(
                        Entity, RequestValidationExceptionValueSchema
                    ):
                        final_field = RequestValidationExceptionValueSchema
                    elif issubclass(Entity, Exception):
                        final_field = ErrorValueSchema
                    elif issubclass(Entity, Model):
                        final_field = Entity
                    else:
                        raise TypeError(
                            f"unsupported obj type {type(Entity)}"
                        )
                case _:
                    raise UnsupportedIndicatorError(
                        f"indicator {v} is not supported"
                    )
            schema_kwargs[k] = (final_field, ...)

        try:
            return self.__created_schema_by_name[schema_name]
        except KeyError:
            CreatedModel: type[Model] = Model.create_dynamic(
                schema_name,
                **schema_kwargs
            )
            self.__created_schema_by_name[schema_name] = CreatedModel
            return CreatedModel

    def __find_locations(self) -> _Locations:
        return {
            Indicator.TYPE: find_location_by_field(Indicator.TYPE, self.__mp),
            Indicator.VALUE: find_location_by_field(Indicator.VALUE, self.__mp)
        }

    @property
    def items(self) -> ItemsView[str, Indicator]:
        return self.__mp.items()

    def digest(self, obj: Indicatable) -> dict:
        """Traverses object to create dictionary based on defined
        indicators.

        Args:
            obj:
                Object to be digested.

        Returns:
            Dictionary with keys and values complying this indication.

        Raises:
            UnsupportedIndicatorError:
                Occured indicator is not supported.
        """
        result: dict = {}

        for k, v in self.items:
            final_field: str | dict
            match v:
                case Indicator.TYPE:
                    final_field = self.__get_type_for_object(obj).value
                case Indicator.VALUE:
                    if isinstance(obj, HttpException):
                        final_field = {
                            "status_code": obj.status_code,
                            "message": obj.detail
                        }
                    elif isinstance(obj, RequestValidationException):
                        message: str = ""
                        locations: list[list[str]] = []
                        validation_errors = obj.errors()

                        for err in validation_errors:
                            if err.get("msg", None):
                                # In case of several errors separate their
                                # messages all in one
                                if message:
                                    message += " ;; "
                                message += err["msg"]
                            err_loc: tuple[str | int] = err["loc"]
                            validation.validate_each(
                                err_loc,
                                [str, int],
                                expected_sequence_type=tuple
                            )
                            locations.append(list(validation.apply(
                                err_loc,
                                tuple
                            )))

                        final_field = {
                            "message": message,
                            "locations": locations
                        }
                    elif isinstance(obj, Exception):
                        final_field = {
                            "message": " ;; ".join([str(x) for x in obj.args]),
                            "code": self.__get_final_error_code(
                                obj
                            )
                        }
                    elif isinstance(obj, Model):
                        # Since pydantic seems to not having a way to deal
                        # with Enums through dict() call (only via Config),
                        # here we do not performance good operation of
                        # converting to json and back to ensure that no Python
                        # objects is remained in digested model.
                        final_field = json.loads(obj.json())
                    else:
                        raise TypeError(
                            f"unsupported obj type {type(obj)}"
                        )
                case _:
                    raise UnsupportedIndicatorError(
                        f"indicator {v} is not supported"
                    )
            result[k] = final_field

        return result

    def recover(
        self,
        Object: type[IndicatableTypeVar],
        mp: dict
    ) -> IndicatableTypeVar:
        """Creates object from given map according to indication rules.

        Note that your project's indication instance should match an indication
        instance used to digest the object.

        Only instances of base classes subclasses can be recovered. E.g.
        Model or Error.

        Recovering of Python's Exception are also supported.

        Args:
            Object:
                Object type to recover.
            mp:
                Dictionary from which object will be recovered.

        Returns:
            Recovered object.

        Raises:
            RecoveringError:
                TYPE or VALUE locations of given dictionary are not matched
                with according indication data.
        """
        validate(Object, type)
        validate_dict(mp, (str, Validator.SKIP))

        # Note that in this case availability of Indicator.TYPE field in given
        # map is not checked since only Indicator.VALUE is needed.
        value: dict = self.__find_value_in_mp(mp)

        if (
            issubclass(Object, HttpException)
        ):
            return Object(
                status_code=value["status_code"],
                detail=value["message"]
            )
        elif (
            issubclass(Object, RequestValidationException)
        ):
            # Temporarily RequestValidationException data is not recovered
            return Object(
                errors=[]
            )
        elif (
            issubclass(Object, Exception)
        ):
            return Object(*value["message"].split(" ; "))
        elif issubclass(Object, Model):
            return Object.parse_obj(value)
        else:
            raise TypeError(
                f"unrecognized Object {Object}"
            )

    def __find_value_in_mp(self, mp: dict) -> dict:
        indication_value_field_location: FieldLocation = \
            self.__locations[Indicator.VALUE]
        mp_value: dict = find_field_by_location(
            indication_value_field_location, mp
        )

        validate(mp_value, dict)
        return mp_value

    def __get_type_for_object(
        self,
        obj: Indicatable
    ) -> IndicationType:
        indication_type: IndicationType = IndicationType.OK

        # For models default type is OK, unless is defined
        # otherwise in INDICATION_TYPE
        if isinstance(obj, Model) and obj.INDICATION_TYPE is not None:
            indication_type = validation.apply(
                obj.INDICATION_TYPE,
                IndicationType
            )
        # For Exception indication type is always ERROR
        elif isinstance(obj, Exception):
            indication_type = IndicationType.ERROR

        return indication_type

    def __get_final_error_code(self, err: Exception) -> int | str | None:
        try:
            error_code: ErrorCode = get_error_code(err)
        except AttributeError:
            return None

        final_error_code: int | str

        if isinstance(error_code, Enum):
            enum_value: Any = error_code.value
            try:
                validation.validate(
                    enum_value,
                    [int, str]
                )
            except validation.ValidationError as err:
                raise validation.ValidationError(
                    f"enum {error_code} contains unsupported by error codes"
                    f" type: {type(enum_value)}, only str or int are"
                    " supported"
                )

            final_error_code = enum_value

        else:
            if (
                not isinstance(error_code, int)
                and not isinstance(error_code, str)
            ):
                raise validation.ValidationError(
                    "error code should be int, str or enum,"
                    f" got {type(error_code)}"
                )
            final_error_code = error_code

        return final_error_code
